use crate::{NodeKind, ParseError, ParseEvent, Parser};

/// Abtract, lossy representation of the syntax.
#[derive(Debug)]
pub struct Code<'a>(pub Vec<Statement<'a>>);

#[derive(Debug)]
pub enum Statement<'a> {
    Pipeline(Pipeline<'a>),
}

#[derive(Debug)]
pub struct Pipeline<'a>(pub Vec<Command<'a>>);

#[derive(Debug, Clone, Copy)]
pub enum Redirect {
    None,
    Stdout,
}

#[derive(Debug)]
pub struct Command<'a>(pub Vec<Word<'a>>, pub Redirect);

#[derive(Debug)]
pub struct Word<'a>(pub Vec<WordPart<'a>>);

#[derive(Debug)]
pub enum WordPart<'a> {
    Text(&'a str),
}

impl<'a> TryFrom<&'a str> for Code<'a> {
    type Error = ParseError;

    fn try_from(source: &'a str) -> Result<Self, Self::Error> {
        let mut parser = Parser::new(source);
        let mut pos = 0;
        build_tree_program(&mut parser, source, &mut pos)
    }
}

fn build_tree_program<'a>(
    parser: &mut Parser,
    source: &'a str,
    pos: &mut usize,
) -> Result<Code<'a>, ParseError> {
    let mut children = vec![];

    while let Some(ev) = parser.next() {
        match ev {
            ParseEvent::StartNode(NodeKind::Pipeline) => children.push(Statement::Pipeline(
                build_tree_pipeline(parser, source, pos)?,
            )),
            ParseEvent::EndNode => break,
            ParseEvent::NewLeaf(_, len) => *pos += len,
            ParseEvent::Error(err, _) => return Err(err),
            _ => unreachable!(),
        }
    }

    Ok(Code(children))
}

fn build_tree_pipeline<'a>(
    parser: &mut Parser,
    source: &'a str,
    pos: &mut usize,
) -> Result<Pipeline<'a>, ParseError> {
    let mut children = vec![];

    while let Some(ev) = parser.next() {
        match ev {
            ParseEvent::StartNode(NodeKind::Command) => {
                children.push(build_tree_command(parser, source, pos)?)
            }
            ParseEvent::NewLeaf(NodeKind::Pipe, _) => {
                *pos += 1;
                children.last_mut().unwrap().1 = Redirect::Stdout;
            }
            ParseEvent::EndNode => break,
            ParseEvent::NewLeaf(_, len) => *pos += len,
            ParseEvent::Error(err, _) => return Err(err),
            _ => unreachable!(),
        }
    }

    Ok(Pipeline(children))
}

fn build_tree_command<'a>(
    parser: &mut Parser,
    source: &'a str,
    pos: &mut usize,
) -> Result<Command<'a>, ParseError> {
    let mut children = vec![];

    while let Some(ev) = parser.next() {
        match ev {
            ParseEvent::StartNode(NodeKind::Word) => {
                children.push(build_tree_word(parser, source, pos)?)
            }
            ParseEvent::EndNode => break,
            ParseEvent::NewLeaf(_, len) => *pos += len,
            ParseEvent::Error(err, _) => return Err(err),
            _ => unreachable!(),
        }
    }

    Ok(Command(children, Redirect::None))
}

fn build_tree_word<'a>(
    parser: &mut Parser,
    source: &'a str,
    pos: &mut usize,
) -> Result<Word<'a>, ParseError> {
    let mut children = vec![];

    while let Some(ev) = parser.next() {
        match ev {
            ParseEvent::NewLeaf(NodeKind::PlainText, len) => {
                children.push(WordPart::Text(&source[*pos..*pos + len]));
                *pos += len
            }
            ParseEvent::NewLeaf(NodeKind::EscapedChar, len) => {
                children.push(WordPart::Text(&source[*pos + 1..*pos + len]));
                *pos += len
            }
            ParseEvent::StartNode(NodeKind::DQuotedString) => {
                build_tree_string(parser, source, pos, &mut children)?
            }
            ParseEvent::EndNode => break,
            ParseEvent::NewLeaf(_, len) => *pos += len,
            ParseEvent::Error(err, _) => return Err(err),
            _ => unreachable!(),
        }
    }

    Ok(Word(children))
}

fn build_tree_string<'a>(
    parser: &mut Parser,
    source: &'a str,
    pos: &mut usize,
    children: &mut Vec<WordPart<'a>>,
) -> Result<(), ParseError> {
    for ev in parser {
        match ev {
            ParseEvent::NewLeaf(NodeKind::PlainText, len) => {
                children.push(WordPart::Text(&source[*pos..*pos + len]));
                *pos += len
            }
            ParseEvent::NewLeaf(NodeKind::EscapedChar, len) => {
                children.push(WordPart::Text(&source[*pos + 1..*pos + len]));
                *pos += len
            }
            ParseEvent::EndNode => break,
            ParseEvent::NewLeaf(_, len) => *pos += len,
            ParseEvent::Error(err, _) => return Err(err),
            _ => unreachable!(),
        }
    }

    Ok(())
}
