//! The runtime for executing oyster programs.
//! Panics when an invalid ast gets passed.

pub mod builtins;
mod pipeline;

use std::io;

use builtins::BuiltinMap;
use nix::{
    errno::Errno,
    sys::{
        signal::{self, SaFlags, SigAction, SigHandler, Signal},
        signalfd::SigSet,
    },
};
use oyster_parser::ast;
use thiserror::Error;

/// Exit status of a finished command.
#[derive(Debug)]
pub struct Status(pub i32);

impl Status {
    pub const SUCCESS: Status = Status(0);
    pub const COULD_NOT_EXEC: Status = Status(126);
    pub const COMMAND_NOT_FOUND: Status = Status(127);
    pub const SIG_BASE: Status = Status(128);
}

/// Errors that occur during runtime.
#[derive(Debug, Error)]
pub enum RuntimeError {
    #[error("failed to create pipe: {0}")]
    PipeCreationFailed(#[source] nix::Error),

    #[error("command not found: {0}")]
    CommandNotFound(String),

    #[error("permission denied: {0}")]
    PermissionDenied(String),

    #[error("failed to spawn process: {0}")]
    SpawnFailed(#[source] io::Error),

    #[error("failed to fork: {0}")]
    ForkFailed(#[source] Errno),

    #[error("waitpid error: {0}")]
    WaidPid(nix::Error),
}

pub struct Shell {
    builtins: BuiltinMap,
}

impl Shell {
    pub fn new() -> Result<Shell, nix::Error> {
        let ignore = SigAction::new(SigHandler::SigIgn, SaFlags::empty(), SigSet::empty());

        unsafe {
            signal::sigaction(Signal::SIGTSTP, &ignore)?;
            signal::sigaction(Signal::SIGTTOU, &ignore)?;
        }

        Ok(Shell {
            builtins: Default::default(),
        })
    }

    pub fn run<'a>(&mut self, code: &'a ast::Code) -> Result<Status, RuntimeError> {
        let mut last_status = Status::SUCCESS;

        for stmt in code.0.iter() {
            last_status = match stmt {
                ast::Statement::Pipeline(pipeline) => self.run_pipeline(pipeline)?,
            };
        }

        Ok(last_status)
    }
}
